/*
  Copyright 2006 Patrik Nilsson and Daniel Nilsson

  This file is part of shpirn.
  
  shpirn is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
	
  shpirn is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with shpirn; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "shpirn.h"
#include "lpio.h"

int keep_log;
int portnum;

void signal_handler( int num );

int main( int argc, char **argv )
{
  static struct option long_options[] =
  {
	{ "help", no_argument, 0, 'h' },
	{ "test", no_argument, 0, 't' },
	{ "logfile", required_argument, 0, 'l' },
	{ "port", required_argument, 0, 'p' },
	{ 0, 0, 0, 0 }
  };

  int choice;
  int opt_index; /* getopt_long stores stuff here */
  int status;

  /* register some signals */
  signal( SIGINT, signal_handler );
  signal( SIGTERM, signal_handler );

  /* be nice and present our self */
  printf( "shpirn v%s\n", VERSION );
#ifdef DEBUG
  printf( "debug code is active...\n" );
#endif
  
  /* set some defaults */
  keep_log = FALSE;
  portnum = NETPORT;

  /* check opts */
  while( 1 )
  {
    opt_index = 0;

    choice = getopt_long( argc, argv, "htl:p:", long_options, &opt_index );
    
	if( choice == -1 ) /* if we got no more args */
	  break;

	switch( choice )
    {
	  case 'h':
		printf( "Usage: shpirn [OPTION]...\n" );
		printf( "-some description-\n\n" );
		printf( "  -h, --help\tdisplay this help and exit\n" );
		printf( "  -t, --test\ttests the inputs and exit (hardware testing)\n" );
		printf( "  -l, --logfile\tenables logging, and appends to specified file\n" );
		printf( "  -p, --port\tuse port instead of default (%i)\n", NETPORT );
		exit( 0 );
		break;

	  case 't':
		test_io();
		break;

	  case 'l':
		status = logger_init( optarg );
		keep_log = TRUE;
		if( status == FALSE )
		{
		  fprintf( stderr, "Error: unable to open '%s' for appending\n", optarg );
		  exit( 1 );
		}
		break;
	  case 'p':
		portnum = atoi( optarg );
		break;
	}
  }
  /* do yer other stuff */
  status = real_loop();
  if( keep_log )
	logger_close();
  exit( status );
}

void test_io( void )
{
  /* Test mode */
  int x;
    
  if( lpio_prepare() )
  {
	printf( "Inputs:\n" );
	for( x = 0; x < 8; x++ )
	  printf( "\t%d: %d\n", x, lpio_read( x ) );
	exit( 0 );
  }
  exit( 1 );
}

static int clock;

void signal_handler( int num )
{
   switch( num )
   {
     case SIGALRM:
	   clock++;
	   break;
     case SIGINT:
     case SIGTERM:
	printf( "%s: hit me, exiting nicely\n", sys_siglist[ num ] );
	if( keep_log )
	   logger_close();
	exit( EXIT_SUCCESS );
	break;
   }
}

void init_clock(void)
{
   struct itimerval ck;

   clock = 0;
   memset( &ck, 0, sizeof( ck ) );
   signal( SIGALRM, signal_handler );
   ck.it_interval.tv_usec = 1000000 / CPS;
   ck.it_value.tv_usec = 1000000 / CPS;
   if( setitimer( ITIMER_REAL, &ck, NULL ) == -1 )
   {
	 fprintf( stderr, "Error: %s\n", strerror( errno ) );
	 exit( EXIT_FAILURE );
   }
}

int real_loop( void )
{
  /* This is the io loop etc */
  int pins[8];
  int x, tmp;
  fd_set rfds;
  struct timeval tv;
  int fd_count, hfd;
  int sockfd, on = 1;
  struct sockaddr_in serv_addr;
  Client *client;

  /* init */
  if( !lpio_prepare() )
     return 1;

  for( x = 0; x < 8; x++ )
     pins[x] = lpio_read( x );

  /* network init */
  sockfd = socket( PF_INET, SOCK_STREAM, 0);
  if( sockfd < 0 )
  {
	fprintf( stderr, "Error: opening socket\n" );
	exit( EXIT_FAILURE );
  }
  
  memset( (char *) &serv_addr, 0, sizeof( serv_addr ) );
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = INADDR_ANY;
  serv_addr.sin_port = htons( portnum );
  setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof( on ) );
  if ( bind( sockfd, (struct sockaddr *) &serv_addr, sizeof( serv_addr )) < 0)
  {
	fprintf( stderr, "Error: binding\n" );
	exit( EXIT_FAILURE );
  }
  listen( sockfd, 5 );

  /* start clock */
  init_clock();

  for( ;; )
  {
     FD_ZERO( &rfds );

     tv.tv_sec = 20;
     tv.tv_usec = 0;

     hfd = sockfd + 1;
     FD_SET( sockfd, &rfds );
     foreach_client( client )
     {
       FD_SET( client->fd, &rfds );
       if( client->fd >= hfd )
	 hfd = client->fd + 1;
     }

     fd_count = select( hfd, &rfds, NULL, NULL, &tv );
     if( fd_count == -1 )
     {
	   if( errno != EINTR ) /* signal received */
		 return 1;
     }

     if( fd_count > 0 && FD_ISSET( sockfd, &rfds ) )
     {
       int newfd;
       fd_count--;
       newfd = accept( sockfd, NULL, 0 );
       if( newfd == -1 )
	   {
		 perror( "accept" );
       }
       else
       {
	         client = add_client( newfd );
		 printf( "client connected\n" );
		 /* welcome the client by sendning the current status */
		 sock_send( client->fd, "c:i:" );
		 for( x = 0; x < 7; x++ )
		 {
		   sock_send( client->fd, "%d,", pins[x] );
		 }
		 sock_send( client->fd, "%d\n", pins[7] );
       }
     }
     foreach_client( client )
     {
       if( fd_count > 0 && FD_ISSET( client->fd, &rfds ) )
       {
		 unsigned char data[4096];
		 int rs;
		 fd_count--;
		 rs = read( client->fd, data, sizeof( data ) );
		 if( !rs )
		 {
		   /* lost connection */
		   del_client( client );
		   printf( "client lost\n" );
		   break;
		 }
		 /* handle input from client */
       }
     }

     if( clock > 0 )
	 {
	   for( x = 0; x < 8; x++ )
	   {
		 tmp = lpio_read( x );
		 if( tmp != pins[x] )
		 {
		   pins[x] = tmp;
		   printf( "e:i:%d=%d\n", x, tmp );
		   if( keep_log )
			 logger_write( "e:i:%d=%d\n", x, tmp );
		   /* write to all connected client */
		   foreach_client( client )
			 sock_send( client->fd, "e:i:%d=%d\n", x, tmp );
		 }
	   }
	   clock = 0;
     }
  }

  /* Disconnect all clients */
  while( first_client )
     del_client( first_client );

  return( 0 );
}

void sock_send( int fd, char *fmt, ... )
{
  /* output to fd in a similar fashion as printf */
  va_list ap;
  char *msg = (char *)malloc(MSG_LEN);
  int msglen;

  va_start( ap, fmt );
  msglen = vsnprintf( msg, MSG_LEN, fmt, ap );
  va_end( ap );

  if( write( fd, msg, msglen ) == -1 )
  {
	fprintf( stderr, "Error: %s\n", strerror( errno ) );
	exit( EXIT_FAILURE );
  }
}
